/*
    CFS0.11 filesystem implementation for FUSE
    Copyright (C) 2005-2020 Kajtar Zsolt (Soci/Singular) <soci@c64.rulez.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/
#include "buffer.h"
#include <string.h>

#define CACHE_SIZE 32
static Buffer buffers[CACHE_SIZE];
static int buffer_head, buffer_tail;

void buffers_init(void) {
    int i;
    for (i = 0; i < CACHE_SIZE; i++) {
        buffers[i].offset = -1;
        buffers[i].next = i + 1;
    }
    buffer_head = 0;
}

static Buffer *buffer_find(Offset ide) {
    int i, j = buffer_head;
    Buffer *buffer;
    for (i = 0; i < CACHE_SIZE; i++) {
        buffer = &buffers[j];
        if (buffer->offset == ide) {
            if (j == buffer_head) break;
            buffers[buffer_tail].next = buffer->next;
            buffer->next = buffer_head;
            buffer_head = j;
            break;
        }
        buffer_tail = j;
        j = buffer->next;
    }
    return buffer;
}

int buffer_flush(Buffer *buffer) {
    int ret = 0;
    if (buffer->dirty) {
        ret = block_write(&buffer->block, buffer->offset);
        buffer->dirty = 0;
    }
    return ret;
}

Buffer *buffer_write(Offset ide) {
    Buffer *buffer = buffer_find(ide);
    if (buffer->offset != ide) {
        buffer_flush(buffer);
        buffer->next = buffer_head;
        buffer_head = buffer_tail;
        buffer->offset = ide;
    }
    buffer->fault = 0;
    buffer->dirty = 1;
    return buffer;
}

Buffer *buffer_write_once(Offset ide) {
    Buffer *buffer = buffer_find(ide);
    if (buffer->offset != ide) {
        buffer_flush(buffer);
        buffer->offset = ide;
    }
    buffer->fault = 0;
    buffer->dirty = 1;
    return buffer;
}

Buffer *buffer_read(Offset ide) {
    Buffer *buffer = buffer_find(ide);
    if (buffer->offset != ide) {
        buffer_flush(buffer);
        buffer->next = buffer_head;
        buffer_head = buffer_tail;
        buffer->fault = block_read(&buffer->block, ide);
        buffer->offset = ide;
    }
    return buffer;
}

Buffer *buffer_read_once(Offset ide) {
    Buffer *buffer = buffer_find(ide);
    if (buffer->offset != ide) {
        buffer_flush(buffer);
        buffer->fault = block_read(&buffer->block, ide);
        buffer->offset = ide;
    }
    return buffer;
}

void buffer_done(Buffer *buffer) {
    (void) buffer;
}

int buffers_sync(int ret) {
    int i, j = 0, ret2;
    int resort[CACHE_SIZE];
    for (i = 0; i < CACHE_SIZE; i++) {
        if (buffers[i].dirty) {
            resort[j++] = i;
        }
    }
    while (j--) {
        Buffer *buffer = &buffers[resort[j]];
        ret2 = block_write(&buffer->block, buffer->offset);
        if (ret2) ret = ret2;
        buffer->dirty = 0;
    }
    return ret;
}

